from sklearn.model_selection import ParameterGrid
import ray
from collections import defaultdict
from ..parameters import build_parameters_from_axis_levels
from ..data_generation import DataGeneratingProcessSampler
import numpy as np


def absolute_mean_error(estimate_vals, true_vals):
    non_zeros = np.logical_not(np.isclose(true_vals, 0))
    return 100*np.abs(
        np.mean(
            (estimate_vals[non_zeros] - true_vals[non_zeros])/true_vals[non_zeros]))

ACCURACY_METRICS = {
    "absolute mean bias %": absolute_mean_error,
    "root mean squared error": lambda estimate_vals, true_vals: np.sqrt(
        np.mean((estimate_vals - true_vals)**2))
}

def _sample_dgp(dgp_sampler):
    return dgp_sampler.sample_dgp()

_sample_dgp_remote = ray.remote(_sample_dgp).remote

def _sample_data(dgp):
    # Sample data
    dataset = dgp.generate_dataset()
    return dataset

_sample_data_remote = ray.remote(_sample_data).remote

def _fit_and_apply_model(model_class, estimand, dataset):
    model = model_class(dataset)
    model.fit()
    estimate_val = model.estimate(estimand=estimand)
    true_val = dataset.ground_truth(estimand=estimand)

    return estimate_val, true_val

_fit_and_apply_model_remote = ray.remote(_fit_and_apply_model).remote

def _process_effect_data(sample_effect_data, ray_enabled):
    # Process potentially async results.
    if ray_enabled:
        sample_effect_data = ray.get(sample_effect_data)

    # Extract estimane and ground truth
    sample_effect_data = np.array(sample_effect_data)
    estimate_vals = sample_effect_data[:, 0]
    true_vals = sample_effect_data[:, 1]

    return estimate_vals, true_vals

def run_concrete_dgp_benchmark(
    dgp, model_class, estimand, num_samples_from_dgp,
    enable_ray_multiprocessing=False):

    # Configure multiprocessing
    if enable_ray_multiprocessing:
        if not ray.is_initialized():
            ray.init()

        sample_data = _sample_data_remote
        fit_and_apply_model = _fit_and_apply_model_remote
    else:
        sample_data = _sample_data
        fit_and_apply_model = _fit_and_apply_model

    sample_effect_data = []
    for _ in range(num_samples_from_dgp):
        dataset = sample_data(dgp)

        # Fit model and use to generate estimates.
        effect_data = fit_and_apply_model(
            model_class, estimand, dataset)
        sample_effect_data.append(effect_data)

    # Extract estimates and ground truth results.
    estimate_vals, true_vals = _process_effect_data(
        sample_effect_data, ray_enabled=enable_ray_multiprocessing)

    results = {}
    for metric_name, metric_func in ACCURACY_METRICS.items():
        results[metric_name] = metric_func(estimate_vals, true_vals)

    return results

def run_sampled_dgp_benchmark(
    dgp_class,
    model_class, estimand,
    data_source_generator, param_grid,
    num_dgp_samples=1,
    num_data_samples_per_dgp=1,
    dgp_kwargs={},
    param_overrides={},
    enable_ray_multiprocessing=False):

    # Configure multiprocessing
    if enable_ray_multiprocessing:
        if not ray.is_initialized():
            ray.init()

        sample_dgp = _sample_dgp_remote
    else:
        sample_dgp = _sample_dgp

    results = defaultdict(list)

    # Iterate over all DGP sampler parameter configurations
    for param_spec in ParameterGrid(param_grid):

        # Construct the DGP sampler for these params.
        dgp_params = build_parameters_from_axis_levels(param_spec)

        for param_name in param_overrides:
            dgp_params.set_parameter(param_name, param_overrides[param_name])

        # Sample DGPs
        # sample_effect_data = []
        metric_results = defaultdict(list)
        for _ in range(num_dgp_samples):
            data_source = data_source_generator()

            dgp_sampler = DataGeneratingProcessSampler(
                dgp_class=dgp_class,
                parameters=dgp_params,
                data_source=data_source,
                dgp_kwargs=dgp_kwargs)

            # Sample data from the DGP
            dgp = sample_dgp(dgp_sampler)

            metric_data = run_concrete_dgp_benchmark(
                dgp, model_class, estimand,
                num_samples_from_dgp=num_data_samples_per_dgp,
                enable_ray_multiprocessing=enable_ray_multiprocessing)

            # sample_effect_data.extend(dgp_sample_effect_data)
            for metric_name, _ in ACCURACY_METRICS.items():
                metric_results[metric_name].append(metric_data[metric_name])

            metric_results["eqn"].append(dgp.treatment_assignment_logit_function)

        # Store the params for this run in the results dict
        for param_name, param_value in param_spec.items():
            results[f"param_{param_name.lower()}"].append(param_value)

        # Calculate and store the requested metric values.
        for metric_name, _ in ACCURACY_METRICS.items():
            results[metric_name] = np.mean(metric_results[metric_name])
            results[metric_name + " (std)"] = np.std(metric_results[metric_name])

    return results, metric_results



from sympy.abc import a, c, x, y, z
import sympy as sp
from pkg_resources import resource_filename
import yaml

### Define constants ###

# OPERATIONAL CONSTANTS
class Constants:
    '''
    This class defines constants which are used throughout
    the package.
    '''

    class Params:
        _SCHEMA_PATH = resource_filename(
            'maccabee', 'parameters/parameter_schema.yml')

        with open(_SCHEMA_PATH, "r") as schema_file:
            SCHEMA = yaml.safe_load(schema_file)["SCHEMA"]

        DEFAULT_SPEC_PATH = resource_filename(
            'maccabee', 'parameters/default_parameter_specification.yml')
        METRIC_LEVEL_SPEC_PATH = resource_filename(
            'maccabee', 'parameters/metric_level_parameter_specifications.yml')

        class ParamInfo:
            DESCRIPTION_KEY = "description"

            TYPE_KEY = "type"
            TYPE_NUMBER = "number"
            TYPE_BOOL = "bool"
            TYPE_DICTIONARY = "dictionary"
            TYPE_CALCULATED = "calculated"

            # Number type keys
            MIN_KEY = "min"
            MAX_KEY = "max"

            # Dict type keys
            DICT_KEYS_KEY = "required_keys"

            # Calc type keys
            EXPRESSION_KEY = "expr"

    # TODO: Move all of the below into a namespace equivalent to
    # other constants.

    LINEAR = "LINEAR"
    POLY_QUADRATIC = "POLY_QUAD"
    POLY_CUBIC = "POLY_CUBIC"
    STEP_CONSTANT = "STEP_JUMP"
    STEP_VARIABLE = "STEP_KINK"
    INTERACTION_TWO_WAY = "INTERACTION_TWO_WAY"
    INTERACTION_THREE_WAY = "INTERACTION_THREE_WAY"

    COVARIATE_SYMBOLS_KEY = "covariates"
    EXPRESSION_KEY = "expr"
    DISCRETE_ALLOWED_KEY = "disc"

    SUBFUNCTION_CONSTANT_SYMBOLS = {a, c}

    # The various transforms which can be applied to covariates
    # and combinations of covariates.
    SUBFUNCTION_FORMS = {
        LINEAR: {
                COVARIATE_SYMBOLS_KEY: [x],
                EXPRESSION_KEY: c*x,
                DISCRETE_ALLOWED_KEY: True
        },
        POLY_QUADRATIC: {
                COVARIATE_SYMBOLS_KEY: [x],
                EXPRESSION_KEY: c*(x**2),
                DISCRETE_ALLOWED_KEY: False
        },
        POLY_CUBIC: {
                COVARIATE_SYMBOLS_KEY: [x],
                EXPRESSION_KEY: c*(x**3),
                DISCRETE_ALLOWED_KEY: False
        },
        STEP_CONSTANT: {
                COVARIATE_SYMBOLS_KEY: [x],
                EXPRESSION_KEY: sp.Piecewise((0, x < a), (c, True)),
                DISCRETE_ALLOWED_KEY: False
        },
        STEP_VARIABLE: {
                COVARIATE_SYMBOLS_KEY: [x],
                EXPRESSION_KEY: sp.Piecewise((0, x < a), (c*x, True)),
                DISCRETE_ALLOWED_KEY: False
        },
        INTERACTION_TWO_WAY: {
                COVARIATE_SYMBOLS_KEY: [x, y],
                EXPRESSION_KEY: c*x*y,
                DISCRETE_ALLOWED_KEY: True
        },
        INTERACTION_THREE_WAY: {
                COVARIATE_SYMBOLS_KEY: [x, y, z],
                EXPRESSION_KEY: c*x*y*z,
                DISCRETE_ALLOWED_KEY: True
        },
    }

    # All X times the number of original covariates.
    MAX_RATIO_TRANSFORMED_TO_ORIGINAL_TERMS = 5

    # How much of the covariate data to use when normalizing functions.
    NORMALIZATION_DATA_SAMPLE_FRACTION = 1

    COVARIATES_NAME = "X"
    TRANSFORMED_COVARIATES_NAME = "TRANSFORMED_X"

    PROPENSITY_LOGIT_NAME = "logit(P(T|X))"
    PROPENSITY_SCORE_NAME = "P(T|X)"
    TREATMENT_ASSIGNMENT_NAME = "T"
    TREATMENT_ASSIGNMENT_SYMBOL = sp.symbols(TREATMENT_ASSIGNMENT_NAME)

    OBSERVED_OUTCOME_NAME = "Y"
    POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME = "Y0"
    POTENTIAL_OUTCOME_WITH_TREATMENT_NAME = "Y1"
    TREATMENT_EFFECT_NAME = "TE"

    OUTCOME_NOISE_NAME = "NOISE(Y)"
    OUTCOME_NOISE_SYMBOL = sp.symbols(OUTCOME_NOISE_NAME)

    ### Metric constants ###

    # Data inputs
    class AnalysisMetricData:
        OBSERVED_COVARIATE_DATA = "OBSERVED_COVARIATES"
        OBSERVED_OUTCOME_DATA = "OBSERVED_OUTCOMES"
        TRANSFORMED_COVARIATE_DATA = "TRANSFORMED_COVARIATES"
        ORACLE_OUTCOME_DATA = "ORACLE_OUTCOMES"

    # Functions
    class AnalysisMetricFunctions:
        LINEAR_R2 = "Lin r2"
        LOGISTIC_R2 = "Log r2"
        PERCENT = "Percent"
        L2_MEAN_DIST = "mean dist"
        NN_CF_MAHALA_DIST = "NN c-factual dist"
        STD_RATIO = "Normed std"
        WASS_DIST = "Wass dist"
        NAIVE_TE = "Naive TE"

    class AxisNames:
        OUTCOME_NONLINEARITY = "OUTCOME_NONLINEARITY"
        TREATMENT_NONLINEARITY = "TREATMENT_NONLINEARITY"
        PERCENT_TREATED = "PERCENT_TREATED"
        OVERLAP = "OVERLAP"
        BALANCE = "BALANCE"
        ALIGNMENT = "ALIGNMENT"
        TE_HETEROGENEITY = "TE_HETEROGENEITY"

    class AxisLevels:
        LOW = "LOW"
        MEDIUM = "MEDIUM"
        HIGH = "HIGH"

    class Data:
        get_dataset_path = lambda file_name: resource_filename(
            'maccabee', f"data/{file_name}.csv")

        LALONDE_PATH = get_dataset_path("lalonde")
        LALONDE_DISCRETE_COVARS = ["black","hispanic","married","nodegree"]

        CPP_PATH = get_dataset_path("cpp")
        CPP_DISCRETE_COVARS = ['x_17','x_22','x_38','x_51','x_54','x_2_A','x_2_B',
            'x_2_C','x_2_D','x_2_E','x_2_F','x_21_A','x_21_B','x_21_C','x_21_D',
            'x_21_E','x_21_F','x_21_G','x_21_H','x_21_I','x_21_J','x_21_K',
            'x_21_L','x_21_M','x_21_N','x_21_O','x_21_P','x_24_A','x_24_B',
            'x_24_C','x_24_D', 'x_24_E']

    class Model:
        ITE_ESTIMAND = "ITE"
        ATE_ESTIMAND = "ATE"



import numpy as np
import sympy as sp
from sympy.abc import x
import yaml
from ..constants import Constants

PARAM_CONSTANTS = Constants.Params

# TODO: refactor the way calculated params and distributions are handled.
class ParameterStore():
    '''
    This class stores all the parameters which are used to control the sampling
    of the Data Generating Process. It ensures that supplied parameters meet
    the fixed schema and is responsible for managing
    calculated and sampling parameters (which are not concretely specified).
    '''

    def __init__(self, parameter_spec_path):
        with open(parameter_spec_path, "r") as params_file:
            raw_parameter_dict = yaml.safe_load(params_file)
        self.parsed_parameter_dict = {}
        self.calculated_parameters = {}

        # Read in the parameter values for each param in the
        # schema.
        for param_name, param_info in PARAM_CONSTANTS.SCHEMA.items():
            param_type = param_info[PARAM_CONSTANTS.ParamInfo.TYPE_KEY]

            # If param should be calculated
            if param_type == PARAM_CONSTANTS.ParamInfo.TYPE_CALCULATED:
                if param_name in raw_parameter_dict:
                    raise Exception(
                        "{} is calculated. It can't be supplied.".format(
                            param_name))

                param_value = self.get_calculated_param_value(param_info)
                self.calculated_parameters[param_name] = param_info

            # If the parameter is in the specification file
            elif param_name in raw_parameter_dict:
                param_value = raw_parameter_dict[param_name]
                if not self.validate_param_value(param_info, param_value):
                    raise Exception("Invalid value for {}: {}".format(
                        param_name, param_value))

            # Parameter is missing.
            else:
                raise Exception("Param spec is missing: {}".format(param_name))

            self.set_parameter(
                param_name, param_value,
                recalculate_calculated_params=False)

    def get_calculated_param_value(self, param_info):
        expr = param_info[PARAM_CONSTANTS.ParamInfo.EXPRESSION_KEY]
        return eval(expr, globals(), self.parsed_parameter_dict)

    def recalculate_calculated_params(self):
        for param_name, param_info in self.calculated_parameters.items():
            param_value = self.get_calculated_param_value(param_info)
            self.set_parameter(param_name, param_value,
                recalculate_calculated_params=False)

    def validate_param_value(self, param_info, param_value):
        param_type = param_info[PARAM_CONSTANTS.ParamInfo.TYPE_KEY]
        if param_type == PARAM_CONSTANTS.ParamInfo.TYPE_NUMBER:
            return param_info[PARAM_CONSTANTS.ParamInfo.MIN_KEY] <= param_value <= param_info[PARAM_CONSTANTS.ParamInfo.MAX_KEY]

        elif param_type == PARAM_CONSTANTS.ParamInfo.TYPE_DICTIONARY:
            required_keys = set(param_info[PARAM_CONSTANTS.ParamInfo.DICT_KEYS_KEY])
            supplied_keys = set(param_value.keys())
            return required_keys == supplied_keys

        elif param_type == PARAM_CONSTANTS.ParamInfo.TYPE_BOOL:
            return param_value in [True, False]

        else:
            # Unknown param type, cannot validate. Fail at medium volume.
            return False

    # Makes a parameter value available on the ParamStore object
    # and stores the value in a dict for later write out.
    def set_parameter(self, param_name, param_value, recalculate_calculated_params=True):
        setattr(self, param_name, param_value)
        self.parsed_parameter_dict[param_name] = param_value
        if recalculate_calculated_params:
            self.recalculate_calculated_params()

    def write(self):
        # TODO: dump parsed params to valid yaml spec
        # with open('data.yml', 'w') as outfile:
        #     yaml.dump(data, outfile, default_flow_style=False)
        pass

    # TODO: provide a wat to specify sampling functions
    # to avoid hard coding.
    def sample_subfunction_constants(self, size=1):
        std = 5*np.sqrt(self.SUBFUNCTION_CONSTANT_TAIL_THICKNESS/(self.SUBFUNCTION_CONSTANT_TAIL_THICKNESS-2))
        return np.round(np.random.standard_t(
                             self.SUBFUNCTION_CONSTANT_TAIL_THICKNESS, size=size)/std, 3)

    def sample_outcome_noise(self, size=1):
        std = 3*np.sqrt(self.OUTCOME_NOISE_TAIL_THICKNESS/(self.OUTCOME_NOISE_TAIL_THICKNESS-2))
        return np.round(np.random.standard_t(
                            self.OUTCOME_NOISE_TAIL_THICKNESS, size=size)/std, 3)

    def sample_treatment_effect(self, size=1):
        return np.round(np.random.standard_t(
                            self.TREATMENT_EFFECT_TAIL_THICKNESS, size=size), 3)

def build_parameters_from_specification(parameter_spec_path):
    '''
    Build a parameter store from a give specification file.
    '''
    return ParameterStore(parameter_spec_path=parameter_spec_path)

def build_parameters_from_axis_levels(metric_levels, save=False):
    '''
    Build a parameter store from a set of metric levels. These
    are applied onto the default parameter spec.
    '''

    params = ParameterStore(parameter_spec_path=PARAM_CONSTANTS.DEFAULT_SPEC_PATH)

    with open(PARAM_CONSTANTS.METRIC_LEVEL_SPEC_PATH, "r") as metric_level_file:
        metric_level_param_specs = yaml.safe_load(metric_level_file)

    # Set the value of each metric to the correct values.
    for metric_name, metric_level in metric_levels.items():

        if metric_name in metric_level_param_specs:
            metric_level_specs = metric_level_param_specs[metric_name]

            if metric_level in metric_level_specs:
                for param_name, param_value in metric_level_specs[metric_level].items():
                    params.set_parameter(
                        param_name, param_value, recalculate_calculated_params=False)
            else:
                raise Exception(f"{metric_level} is not a valid level for {metric_name}")
        else:
            raise Exception(f"{metric_name} is not a valid metric")


    params.recalculate_calculated_params()

    return params



from ..constants import Constants
from .data_set import DataSet
from ..utilities import evaluate_expression
import pandas as pd
import numpy as np
from functools import partial


GENERATED_DATA_DICT_NAME = "_generated_data"

class DataGeneratingMethodClass(type):
    def __init__(cls, name, bases, clsdict):
        for method_name, method_obj in clsdict.items():
            if method_name.startswith('_generate'):
                if type(method_obj) != DataGeneratingMethodWrapper:
                    raise Exception(
                        f"{method_obj} is a _generate* method without the data_generating_method decorator.")

        super(DataGeneratingMethodClass, cls).__init__(name, bases, clsdict)

class DataGeneratingMethodWrapper():
    def __init__(self,
        generated_var, required_vars,
        optional, analysis_mode_only, cache_result,
        func):

        self.generated_var = generated_var
        self.required_vars = required_vars
        self.optional = optional
        self.analysis_mode_only = analysis_mode_only
        self.cache_result = cache_result
        self.func = func

    def __call__(self, *args, **kwargs):
        dgp = args[0]
        data_dict = getattr(dgp, GENERATED_DATA_DICT_NAME)

        if self.cache_result and (self.generated_var in data_dict):
            return data_dict[self.generated_var]

        required_var_vals = {
            k:data_dict[k]
            for k in self.required_vars
            if k in data_dict
        }

        # Have all required inputs.
        if len(required_var_vals) == len(self.required_vars):
            # Only run analysis_mode_only methods if dgp in analysis mode.
            if dgp.analysis_mode or not self.analysis_mode_only:
                val = self.func(args[0], required_var_vals, *args[1:], **kwargs)
                data_dict[self.generated_var] = val
                return val

        # Do not have all required inputs.
        elif not self.optional:
            raise Exception(
                f"Missing required value in non-optional method: {self.func}" )

    def __get__(self, instance, owner):
        return partial(self.__call__, instance)

def data_generating_method(
    generated_var, required_vars,
    optional=False, analysis_mode_only=False, cache_result=False):
    return partial(DataGeneratingMethodWrapper,
        generated_var, required_vars,
        optional, analysis_mode_only, cache_result)

#TODO: consider refctoring to splat in the args directly rather
# than via a dict.

# TODO: consider specifying order via a list of method names
# and then replace _generate validation to just use this list of names.

class DataGeneratingProcess(metaclass=DataGeneratingMethodClass):
    def __init__(self, n_observations, analysis_mode=False):
        setattr(self, GENERATED_DATA_DICT_NAME, {})

        self.n_observations = n_observations
        self.analysis_mode = analysis_mode

    # DGP PROCESS
    def generate_dataset(self):
        # Covars
        self._generate_observed_covars()
        self._generate_transformed_covars()

        # Treatment assignment
        self._generate_true_propensity_scores()
        self._generate_true_propensity_score_logits()
        self._generate_treatment_assignments()

        # Outcomes
        self._generate_outcome_noise_samples()
        self._generate_outcomes_without_treatment()
        self._generate_treatment_effects()
        self._generate_outcomes_with_treatment()
        self._generate_observed_outcomes()

        return self._build_dataset()

    # DGP DEFINITION
    @data_generating_method(Constants.COVARIATES_NAME, [])
    def _generate_observed_covars(self, input_vars):
        raise NotImplementedError

    @data_generating_method(
        Constants.TRANSFORMED_COVARIATES_NAME,
        [Constants.COVARIATES_NAME],
        analysis_mode_only=True)
    def _generate_transformed_covars(self, input_vars):
        return None

    @data_generating_method(
        Constants.PROPENSITY_SCORE_NAME,
        [],
        optional=True)
    def _generate_true_propensity_scores(self, input_vars):
        raise NotImplementedError

    @data_generating_method(
        Constants.PROPENSITY_LOGIT_NAME,
        [Constants.PROPENSITY_SCORE_NAME],
        optional=True,
        analysis_mode_only=True)
    def _generate_true_propensity_score_logits(self, input_vars):
        propensity_scores = input_vars[Constants.PROPENSITY_SCORE_NAME]
        return np.log(propensity_scores/(1-propensity_scores))

    @data_generating_method(
        Constants.TREATMENT_ASSIGNMENT_NAME,
        [Constants.PROPENSITY_SCORE_NAME])
    def _generate_treatment_assignments(self, input_vars):
        propensity_scores = input_vars[Constants.PROPENSITY_SCORE_NAME]
        return (np.random.uniform(
            size=len(propensity_scores)) < propensity_scores).astype(int)


    @data_generating_method(Constants.OUTCOME_NOISE_NAME, [])
    def _generate_outcome_noise_samples(self, input_vars):
        return np.zeros(self.n_observations)

    @data_generating_method(
        Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME,
        [Constants.COVARIATES_NAME])
    def _generate_outcomes_without_treatment(self, input_vars):
        raise NotImplementedError

    @data_generating_method(
        Constants.TREATMENT_EFFECT_NAME,
        [Constants.COVARIATES_NAME])
    def _generate_treatment_effects(self, input_vars):
        raise NotImplementedError

    @data_generating_method(
        Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME,
        [Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME, Constants.TREATMENT_EFFECT_NAME])
    def _generate_outcomes_with_treatment(self, input_vars):
        outcome_without_treatment = input_vars[Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME]
        treatment_effect = input_vars[Constants.TREATMENT_EFFECT_NAME]
        return outcome_without_treatment + treatment_effect

    @data_generating_method(
        Constants.OBSERVED_OUTCOME_NAME,
        [
            Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME,
            Constants.TREATMENT_ASSIGNMENT_NAME,
            Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME,
        ])
    def _generate_observed_outcomes(self, input_vars):
        outcome_without_treatment = input_vars[Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME]
        treatment_assignment = input_vars[Constants.TREATMENT_ASSIGNMENT_NAME]
        outcome_with_treatment = input_vars[Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME]
        return (treatment_assignment*outcome_with_treatment) + ((1-treatment_assignment)*outcome_without_treatment)

    # HELPER FUNCTIONS

    def _get_generated_data(self, name):
        data_dict = getattr(self, GENERATED_DATA_DICT_NAME)
        return data_dict.get(name, None)

    def _build_dataframe_for_vars(self, var_names):
        df = pd.DataFrame()
        for name in var_names:
            val = self._get_generated_data(name)
            df[name] = val

        return df

    def _build_dataset(self):

        # Observed data
        observed_covariate_data = self._get_generated_data(
            Constants.COVARIATES_NAME)

        observed_outcome_data = self._build_dataframe_for_vars([
            Constants.TREATMENT_ASSIGNMENT_NAME,
            Constants.OBSERVED_OUTCOME_NAME
        ])

        # Unobserved data
        transformed_covariate_data = self._get_generated_data(
            Constants.TRANSFORMED_COVARIATES_NAME)

        oracle_outcome_data = self._build_dataframe_for_vars([
            Constants.PROPENSITY_SCORE_NAME,
            Constants.PROPENSITY_LOGIT_NAME,
            Constants.OUTCOME_NOISE_NAME,
            Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME,
            Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME,
            Constants.TREATMENT_EFFECT_NAME,
        ])

        # Treatment assignment
        return DataSet(
            observed_covariate_data=observed_covariate_data,
            observed_outcome_data=observed_outcome_data,
            oracle_outcome_data=oracle_outcome_data,
            transformed_covariate_data=transformed_covariate_data)

class ManualDataGeneratingProcess(DataGeneratingProcess):
    pass

class SampledDataGeneratingProcess(DataGeneratingProcess):
    def __init__(self,
        params,
        observed_covariate_data,
        outcome_covariate_transforms,
        treatment_covariate_transforms,
        treatment_assignment_function,
        treatment_effect_subfunction,
        base_outcome_subfunction,
        treatment_assignment_logit_func=None,
        outcome_function=None,
        analysis_mode=True):

        # STANDARD CONFIG
        n_observations = observed_covariate_data.shape[0]
        super().__init__(n_observations, analysis_mode)

        # SAMPLED SGP CONFIG
        self.params = params

        # Sampled covariate transforms for the treat and outcome functions.
        self.outcome_covariate_transforms = outcome_covariate_transforms
        self.treatment_covariate_transforms = treatment_covariate_transforms

        # Treatment assignment function and subfunctions
        self.treatment_assignment_logit_function = treatment_assignment_logit_func
        self.treatment_assignment_function = treatment_assignment_function

        # Outcome function and subfunctions
        self.treatment_effect_subfunction = treatment_effect_subfunction
        self.base_outcome_subfunction = base_outcome_subfunction
        self.outcome_function = outcome_function

        # DATA
        self.observed_covariate_data = observed_covariate_data

    @data_generating_method(Constants.COVARIATES_NAME, [], cache_result=True)
    def _generate_observed_covars(self, input_vars):
        return self.observed_covariate_data

    @data_generating_method(
        Constants.TRANSFORMED_COVARIATES_NAME,
        [Constants.COVARIATES_NAME],
        analysis_mode_only=True,
        cache_result=True)
    def _generate_transformed_covars(self, input_vars):
        # Generate the values of all the transformed covariates by running the
        # original covariate data through the transforms used in the outcome and
        # treatment functions.

        observed_covariate_data = input_vars[Constants.COVARIATES_NAME]

        all_transforms = list(set(self.outcome_covariate_transforms).union(
            self.treatment_covariate_transforms))

        data = {}
        for index, transform in enumerate(all_transforms):
            data[f"{Constants.TRANSFORMED_COVARIATES_NAME}{index}"] = \
                evaluate_expression(transform, observed_covariate_data)

        return pd.DataFrame(data)


    @data_generating_method(
        Constants.PROPENSITY_SCORE_NAME,
        [Constants.COVARIATES_NAME],
        cache_result=True)
    def _generate_true_propensity_scores(self, input_vars):
        observed_covariate_data = input_vars[Constants.COVARIATES_NAME]

        return evaluate_expression(
            self.treatment_assignment_function,
            observed_covariate_data)

    @data_generating_method(
        Constants.TREATMENT_ASSIGNMENT_NAME,
        [Constants.PROPENSITY_SCORE_NAME])
    def _generate_treatment_assignments(self, input_vars):
        propensity_scores = input_vars[Constants.PROPENSITY_SCORE_NAME]

        # Sample treatment assignment given pre-calculated propensity_scores
        T = (np.random.uniform(
            size=self.n_observations) < propensity_scores).astype(int)

        # Only perform balance adjustment if there is some heterogeneity
        # in the propensity scores.
        if not np.all(np.isclose(propensity_scores, propensity_scores[0])):
            # Balance adjustment
            control_p_scores = propensity_scores.where(T == 0)
            treat_p_scores = propensity_scores.where(T == 1)

            num_controls = control_p_scores.count()
            n_to_switch = int(num_controls*self.params.FORCED_IMBALANCE_ADJUSTMENT)

            control_switch_targets = control_p_scores.nlargest(n_to_switch).index.values
            treat_switch_targets = treat_p_scores.nsmallest(n_to_switch).index.values

            T[control_switch_targets] = 1
            T[treat_switch_targets] = 0

        return T

    @data_generating_method(Constants.OUTCOME_NOISE_NAME, [])
    def _generate_outcome_noise_samples(self, input_vars):
        return self.params.sample_outcome_noise(size=self.n_observations)

    @data_generating_method(
        Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME,
        [Constants.COVARIATES_NAME],
        cache_result=True)
    def _generate_outcomes_without_treatment(self, input_vars):
        observed_covariate_data = input_vars[Constants.COVARIATES_NAME]
        return evaluate_expression(
            self.base_outcome_subfunction,
            observed_covariate_data)

    @data_generating_method(
        Constants.TREATMENT_EFFECT_NAME,
        [Constants.COVARIATES_NAME],
        cache_result=True)
    def _generate_treatment_effects(self, input_vars):
        observed_covariate_data = input_vars[Constants.COVARIATES_NAME]
        return evaluate_expression(
            self.treatment_effect_subfunction,
            observed_covariate_data)

    @data_generating_method(
        Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME,
        [
            Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME,
            Constants.TREATMENT_EFFECT_NAME
        ],
        cache_result=True)
    def _generate_outcomes_with_treatment(self, input_vars):
        outcome_without_treatment = input_vars[Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME]
        treatment_effect = input_vars[Constants.TREATMENT_EFFECT_NAME]
        return outcome_without_treatment + treatment_effect

    @data_generating_method(
        Constants.OBSERVED_OUTCOME_NAME,
        [
            Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME,
            Constants.TREATMENT_ASSIGNMENT_NAME,
            Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME,
            Constants.OUTCOME_NOISE_NAME
        ])
    def _generate_observed_outcomes(self, input_vars):
        outcome_without_treatment = input_vars[Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME]
        treatment_assignment = input_vars[Constants.TREATMENT_ASSIGNMENT_NAME]
        outcome_with_treatment = input_vars[Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME]
        outcome_noise_samples = input_vars[Constants.OUTCOME_NOISE_NAME]
        Y = (
            (treatment_assignment*outcome_with_treatment) +
            ((1-treatment_assignment)*outcome_without_treatment) +
            outcome_noise_samples)

        return Y



import numpy as np
import sympy as sp
from .constants import Constants
import pandas as pd
from functools import partial


def random_covar_matrix(dimension, correlation_deg = 0.5):
    """
    Generate random covariance matrix by approximating the random
    vine method: https://stats.stackexchange.com/questions/2746/
    how-to-efficiently-generate-random-positive-semidefinite-correlation-matrices
    """
    if not (0 <= correlation_deg <= 1):
        raise Exception("Invalid correlation_deg. Must be in [0, 1]")

    # Small K means a larger degree of correlation.
    k = min(max(1, int(dimension*(1-correlation_deg))), dimension-1)
    W = np.random.normal(loc=0, scale=1, size=(dimension, k))
    S = W@W.T + np.diag(np.random.random(dimension))
    S = np.diag(1./np.sqrt(np.diag(S))) @ S @ np.diag(1./np.sqrt(np.diag(S)))

    return S

def extract_treat_and_control_data(covariates, treatment_status):
    X_treated = covariates[(treatment_status==1).to_numpy()]
    X_control = covariates[(treatment_status==0).to_numpy()]
    return X_treated, X_control

def select_given_probability_distribution(full_list, selection_probabilities):
    full_list = np.array(full_list)

    flat = len(full_list.shape) == 1
    if flat:
        full_list = full_list.reshape((-1, 1))

    selections = np.random.uniform(size=full_list.shape[0]) < selection_probabilities
    selected = full_list[selections, :]
    if flat:
        selected = selected.flatten()
    return selected, selections

def evaluate_expression(expression, data):
    try:
        free_symbols = list(expression.free_symbols)

        expr_func = sp.lambdify(
                free_symbols,
                expression,
                modules=[
                    {
                        "amax": lambda x: np.maximum(*x),
                        "amin": lambda x: np.minimum(*x)
                    },
                    "numpy"
                ],
                dummify=True)

        column_data = [data[str(sym)] for sym in free_symbols]
        res = expr_func(*column_data)

        return res
    except AttributeError:
        # No free symbols, return expression itself.
        return expression

def initialize_expression_constants(parameters, expressions):
    initialized_expressions = []

    for expression in expressions:
        constants_to_initialize = \
            Constants.SUBFUNCTION_CONSTANT_SYMBOLS.intersection(expression.free_symbols)

        initialized_expressions.append(expression.subs(
            zip(constants_to_initialize,
                parameters.sample_subfunction_constants(
                    size=len(constants_to_initialize)))))

    return initialized_expressions



from ..constants import Constants
import numpy as np
import pandas as pd


class DataSet():
    '''
    Philosophy: this is an external facing object, so we switch
    to easily usable names and a nice API.
    '''
    # TODO: write tooling to go to and from file to support static
    # benchmarking runs in future.

    # TODO: write tooling for convenient creation of DataSet objects
    # from standard data frames.

    def __init__(self,
        observed_covariate_data,
        observed_outcome_data,
        oracle_outcome_data,
        transformed_covariate_data=None):

        self.observed_covariate_data = observed_covariate_data
        self.observed_outcome_data = observed_outcome_data
        self.oracle_outcome_data = oracle_outcome_data
        self.transformed_covariate_data = transformed_covariate_data

    @property
    def X(self):
        return self.observed_covariate_data

    @property
    def T(self):
        return self.observed_outcome_data[Constants.TREATMENT_ASSIGNMENT_NAME]

    @property
    def Y(self):
        return self.observed_outcome_data[Constants.OBSERVED_OUTCOME_NAME]

    @property
    def Y0(self):
        return self.oracle_outcome_data[
            Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME]

    @property
    def Y1(self):
        return self.oracle_outcome_data[
            Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME]

    @property
    def observed_data(self):
        '''
        Assemble and return the observable data.
        '''

        return self.observed_covariate_data.join(self.observed_outcome_data)

    # Estimand accessors

    @property
    def ITE(self):
        return self.Y1 - self.Y0

    @property
    def ATE(self):
        return np.mean(self.ITE)

    def ground_truth(self, estimand):
        if estimand == Constants.Model.ATE_ESTIMAND:
            return self.ATE
        elif estimand == Constants.Model.ITE_ESTIMAND:
            return self.ITE
        else:
            raise Exception("Unrecognized estimand.")



from sympy.abc import x
import sympy as sp
import numpy as np
import pandas as pd
from itertools import combinations
from ..constants import Constants
from ..utilities import select_given_probability_distribution, evaluate_expression, initialize_expression_constants
from .data_generating_process import SampledDataGeneratingProcess

class DataGeneratingProcessSampler():
    def __init__(self, parameters, data_source,
        dgp_class=SampledDataGeneratingProcess, dgp_kwargs={}):

        self.dgp_class = dgp_class
        self.params = parameters
        self.data_source = data_source
        self.dgp_kwargs = dgp_kwargs

    def sample_dgp(self):
        source_covariate_data = self.data_source.get_data()
        covariate_symbols = np.array(sp.symbols(list(source_covariate_data.columns)))

        # Sample the source data to generate the observed covariate data.
        observed_covariate_data = self.sample_observed_covariate_data(
            source_covariate_data)

        # Select the observed variables which may appear in the assignment or
        # outcome functions. These are potential confounders.
        potential_confounder_symbols = self.sample_potential_confounders(
            covariate_symbols)

        # Sample the covariate transforms which make up the assignment and
        # outcome functions.
        outcome_covariate_transforms, treatment_covariate_transforms = \
            self.sample_treatment_and_outcome_covariate_transforms(
                potential_confounder_symbols)

        treatment_assignment_logit_func, treatment_assignment_function = \
            self.sample_treatment_assignment_function(
                treatment_covariate_transforms, observed_covariate_data)

        outcome_function, base_outcome_subfunc, treat_effect_subfunc = \
            self.sample_outcome_function(
                outcome_covariate_transforms, observed_covariate_data)

        dgp = self.dgp_class(
            params=self.params,
            observed_covariate_data=observed_covariate_data,
            outcome_covariate_transforms=outcome_covariate_transforms,
            treatment_covariate_transforms=treatment_covariate_transforms,
            treatment_assignment_logit_func=treatment_assignment_logit_func,
            treatment_assignment_function=treatment_assignment_function,
            treatment_effect_subfunction=treat_effect_subfunc,
            base_outcome_subfunction=base_outcome_subfunc,
            outcome_function=outcome_function,
            **self.dgp_kwargs)

        return dgp

    def sample_observed_covariate_data(self, source_covariate_data):
        # Sample observations.

        # For now, we allow simple uniform sampling. In future, we may
        # support more complex sampling procedures to allow for simulation
        # of observation censorship etc.

        # NOTE: in the ideal world, this would occur on data generation
        # (as part of the DGP run) rather than here. But normalization of the
        # DGP to target specific distribution properties requires a static
        # dataset.
        observed_covariate_data = source_covariate_data.sample(
            frac=self.params.OBSERVATION_PROBABILITY)

        return observed_covariate_data

    def sample_potential_confounders(self, covariate_symbols):
        potential_confounder_symbols, _ = select_given_probability_distribution(
            full_list=covariate_symbols,
            selection_probabilities=self.params.POTENTIAL_CONFOUNDER_SELECTION_PROBABILITY)

        if len(potential_confounder_symbols) == 0:
            potential_confounder_symbols = [np.random.choice(covariate_symbols)]

        return potential_confounder_symbols

    def sample_covariate_transforms(self, covariate_symbols,
        transform_probabilities, empty_allowed=False):
        """
        Sample a set of transforms which will be applied to the base covariates.
        The set of transforms is governed by the specified probabilities for
        each possible transform type. Each transform is parameterized by
        a set of constants.

        These transforms are used in the finalized functional form of the
        treat/outcome functions. Each of these functions uses
        combines a set of random transforms of the covariates (produced by this
        function). The subfunction form probabilities and combination of the
        transforms is different for the treat and outcome function.
        """
        #TODO NB: Post-process the output of this function to group terms
        # based on the same covariates and produce new multiplicative
        # combinations of different covariates.

        selected_covariate_transforms = []
        for transform_name, transform_spec in Constants.SUBFUNCTION_FORMS.items():
            transform_expression = transform_spec[Constants.EXPRESSION_KEY]
            transform_covariate_symbols = transform_spec[Constants.COVARIATE_SYMBOLS_KEY]
            transform_discrete_allowed = transform_spec[Constants.DISCRETE_ALLOWED_KEY]

            usable_covariate_symbols = covariate_symbols
            if not transform_discrete_allowed:
                usable_covariate_symbols = list(filter(
                    lambda sym: str(sym) not in self.data_source.binary_column_names,
                    covariate_symbols))

            # All possible combinations of covariates for the given transform.
            covariate_combinations = np.array(
                list(combinations(
                    usable_covariate_symbols,
                    len(transform_covariate_symbols))))

            selected_covar_combinations, _ = select_given_probability_distribution(
                full_list=covariate_combinations,
                selection_probabilities=transform_probabilities[transform_name])

            selected_covariate_transforms.extend([transform_expression.subs(
                                    zip(transform_covariate_symbols, covar_comb))
                                    for covar_comb in selected_covar_combinations
                                ])

        # Add at least one transform if empty_allowed is False
        # and no transforms selected above.
        # TODO: this is an ugly solution to a complex problem. Improve this.
        if len(selected_covariate_transforms) == 0 and not empty_allowed:

            # TODO: eval and delete this if no longer desired.
            # transform_spec = Constants.SUBFUNCTION_FORMS[Constants.LINEAR]
            # transform_expression = transform_spec[Constants.EXPRESSION_KEY]
            # transform_covariate_symbols = transform_spec[Constants.COVARIATE_SYMBOLS_KEY]
            # required_covars = covariate_symbols[:len(transform_covariate_symbols)]
            #
            # selected_covariate_transforms = [
            #     transform_expression.subs(zip(
            #     transform_covariate_symbols, required_covars))
            # ]
            selected_covariate_transforms.append(list(Constants.SUBFUNCTION_CONSTANT_SYMBOLS)[0])

        # The number of combinations grows factorially with the number of
        # covariates. Cap the number of covariate transforms at a multiple
        # of the number of base covariates.
        max_transform_count = \
            Constants.MAX_RATIO_TRANSFORMED_TO_ORIGINAL_TERMS*len(covariate_symbols)

        if len(selected_covariate_transforms) > max_transform_count:
            # Randomly sample selected transforms with expected number selected
            # equal to the max.
            selection_p = max_transform_count/len(selected_covariate_transforms)

            selected_covariate_transforms, _ = select_given_probability_distribution(
                full_list=selected_covariate_transforms,
                selection_probabilities=selection_p)

            selected_covariate_transforms = list(selected_covariate_transforms)

        return selected_covariate_transforms

    def sample_treatment_and_outcome_covariate_transforms(self, potential_confounder_symbols):
        """
        Sample covariate transforms for the treatment and outcome function
        and then modify the sampled transforms to generate desired alignment.
        """
        outcome_covariate_transforms = self.sample_covariate_transforms(
                potential_confounder_symbols,
                self.params.OUTCOME_MECHANISM_COVARIATE_SELECTION_PROBABILITY)

        treatment_covariate_transforms = self.sample_covariate_transforms(
                potential_confounder_symbols,
                self.params.TREAT_MECHANISM_COVARIATE_SELECTION_PROBABILITY)

        # Unique set of all covariate transforms
        all_transforms = list(set(
            treatment_covariate_transforms + outcome_covariate_transforms))

        # Select overlapping covariates transforms (effective confounder space)
        # based on the alignment parameter.
        aligned_transforms, _ = select_given_probability_distribution(
                all_transforms,
                selection_probabilities=self.params.ACTUAL_CONFOUNDER_ALIGNMENT)

        treat_only_transforms = list(set(treatment_covariate_transforms).difference(aligned_transforms))
        outcome_only_transforms = list(set(outcome_covariate_transforms).difference(aligned_transforms))

        # Initialize the constants in all the transforms.
        aligned_transforms = initialize_expression_constants(
            self.params,
            aligned_transforms)

        treat_only_transforms = initialize_expression_constants(
            self.params,
            treat_only_transforms)

        outcome_only_transforms = initialize_expression_constants(
            self.params,
            outcome_only_transforms)

        # Union the true confounders into the original covariate selections.
        outcome_covariate_transforms = np.hstack(
            [aligned_transforms, outcome_only_transforms])
        treatment_covariate_transforms = np.hstack(
            [aligned_transforms, treat_only_transforms])

        return outcome_covariate_transforms, treatment_covariate_transforms


    def sample_treatment_assignment_function(self,
        treatment_covariate_transforms, observed_covariate_data):
        """
        Sample a treatment assignment function by combining the sampled covariate
        transforms, initializing the constants, and normalizing the function
        outputs to conform to constraints and targets on the propensity scores.
        """
        # TODO: consider recursively applying the covariate transformation
        # to produce "deep" functions. Probably overkill.
        # TODO: add non-linear activation function
        # TODO: enable overlap adjustment

        # Build base treatment logit function. Additive combination of the true covariates.
        base_treatment_logit_expression = np.sum(
            treatment_covariate_transforms)

        # Sample data to evaluate distribution.
        sampled_data = observed_covariate_data.sample(
            frac=Constants.NORMALIZATION_DATA_SAMPLE_FRACTION)

        # Adjust logit
        logit_values = evaluate_expression(
            base_treatment_logit_expression, sampled_data)

        max_logit = np.max(logit_values)
        min_logit = np.min(logit_values)
        mean_logit = np.mean(logit_values)

        # Rescale to meet min/max constraints and target propensity.
        # First, construct function to rescale between 0 and 1
        diff = (max_logit - min_logit)
        if np.min(diff) > 0:
            normalized_logit_expr = (x - min_logit)/diff
        else:
            prop_score_logit = self.params.TARGET_MEAN_LOGIT
            treatment_assignment_logit_function = prop_score_logit
            treatment_assignment_function = \
                np.exp(prop_score_logit)/(1+np.exp(prop_score_logit))

            return (treatment_assignment_logit_function,
                treatment_assignment_function)

        # Second, construct function to rescale to target interval
        constrained_logit_expr = self.params.TARGET_MIN_LOGIT + \
            (x*(self.params.TARGET_MAX_LOGIT - self.params.TARGET_MIN_LOGIT))

        rescaling_expr = constrained_logit_expr.subs(x, normalized_logit_expr)
        rescaled_logit_mean = rescaling_expr.evalf(subs={x: mean_logit})


        # Third, construct function to apply offset for targeted propensity.
        # This requires the rescaled mean found above.
        target_propensity_adjustment = self.params.TARGET_MEAN_LOGIT - rescaled_logit_mean
        targeted_logit_expr = rescaling_expr + target_propensity_adjustment

        # Apply max/min truncation to account for adjustment shift.
        max_min_capped_targeted_logit = sp.functions.Max(
                sp.functions.Min(targeted_logit_expr, self.params.TARGET_MAX_LOGIT),
                self.params.TARGET_MIN_LOGIT)

        # Finally, construct the full function expression.
        treatment_assignment_logit_function = \
            max_min_capped_targeted_logit.subs(x, base_treatment_logit_expression)

        exponentiated_logit = sp.functions.exp(treatment_assignment_logit_function)
        treatment_assignment_function = exponentiated_logit/(1 + exponentiated_logit)

        return (treatment_assignment_logit_function,
            treatment_assignment_function)

    def sample_treatment_effect_subfunction(self,
        outcome_covariate_transforms, observed_covariate_data):
        """ Create treatment effect subfunction """

        base_treatment_effect = self.params.sample_treatment_effect()[0]

        # Sample outcome subfunctions to interact with Treatment effect.
        selected_interaction_terms, _ = select_given_probability_distribution(
                full_list=outcome_covariate_transforms,
                selection_probabilities=self.params.TREATMENT_EFFECT_HETEROGENEITY)

        initialized_interaction_terms = initialize_expression_constants(
            self.params,
            selected_interaction_terms)

        # Build multiplier
        treatment_effect_multiplier_expr = np.sum(initialized_interaction_terms)

        # Process interaction terms into treatment subfunction.
        if treatment_effect_multiplier_expr != 0:

            # Normalize multiplier size.
            # TODO: vaidate the approach to normalization used here
            # vs other function construction.
            sampled_data = observed_covariate_data.sample(
                frac=Constants.NORMALIZATION_DATA_SAMPLE_FRACTION)

            treatment_effect_multiplier_values = evaluate_expression(
                treatment_effect_multiplier_expr, sampled_data)

            multiplier_std = np.std(treatment_effect_multiplier_values)
            multiplier_mean = np.mean(treatment_effect_multiplier_values)

            normalized_treatment_effect_multiplier_expr = \
                (treatment_effect_multiplier_expr - multiplier_mean)/multiplier_std

            treatment_effect_subfunction = base_treatment_effect * \
                (1+normalized_treatment_effect_multiplier_expr)

            return treatment_effect_subfunction

        # No interaction terms. Return base treatment effect as the
        # treatment effect subfunction.
        else:
            return base_treatment_effect


    def sample_outcome_function(self,
        outcome_covariate_transforms, observed_covariate_data):
        """ Create outcome function"""

        # TODO: consider recursively applying the covariate transformation
        # to produce "deep" functions. Probability overkill.

        # Build base outcome function. Additive combination of the true covariates.

        # TODO: add non-linear activation function and ensure proper normalization.
        # using or changing the OUTCOME_MECHANISM_EXPONENTIATION param.
        base_outcome_expression = np.sum(outcome_covariate_transforms)

        # Sample data to evaluate distribution.
        sampled_data = observed_covariate_data.sample(
            frac=Constants.NORMALIZATION_DATA_SAMPLE_FRACTION)

        # Normalized outcome values to have approximate mean=0 and std=1.
        # This prevents situations where large outcome values drown out
        # the treatment effect or the treatment effect dominates small average outcomes.
        outcome_values = evaluate_expression(base_outcome_expression, sampled_data)
        outcome_mean = np.mean(outcome_values)
        outcome_std = np.std(outcome_values)

        # This is only an approximate normalization. It will shift the mean to zero
        # but the exact effect on std will depend on the distribution.
        normalized_outcome_expression = \
            (base_outcome_expression - outcome_mean)/outcome_std

        # Create the treatment effect subfunction.
        treat_effect_subfunction = self.sample_treatment_effect_subfunction(
            outcome_covariate_transforms, observed_covariate_data)

        base_outcome_subfunction = normalized_outcome_expression
        outcome_function = base_outcome_subfunction + \
            Constants.OUTCOME_NOISE_SYMBOL + \
            (Constants.TREATMENT_ASSIGNMENT_SYMBOL*treat_effect_subfunction)

        return outcome_function, base_outcome_subfunction, treat_effect_subfunction



import numpy as np
import pandas as pd
from .constants import Constants
from .utilities import random_covar_matrix


class DataSource():

    def __init__(self, covariate_data, binary_column_names):
        self.original_covariate_data = covariate_data
        self.binary_column_names = binary_column_names

        self.binary_column_indeces = [
            self.original_covariate_data.columns.get_loc(name)
            for name in self.binary_column_names
        ]

        self.normalized_covariate_data = _normalize_covariate_data(
            self.original_covariate_data,
            exclude_columns=self.binary_column_indeces)

    def get_data(self):
        return self.normalized_covariate_data

    # TODO: Build in a mechanism to limit/control the size of the
    # returned dataset? See below
    # cpp_covars = cpp_covars[[f"x_{i}" for i in range(1, 20)]]
    # Constants.Data.CPP_DISCRETE_COVARS = [i for i in Constants.Data.CPP_DISCRETE_COVARS if i < 20]
    # cpp_covars = cpp_covars.head(200)

def load_cpp():
    cpp_covars = load_covars_from_path(Constants.Data.CPP_PATH)


    cpp_data_source = DataSource(
        covariate_data=cpp_covars,
        binary_column_names=Constants.Data.CPP_DISCRETE_COVARS)

    return cpp_data_source

def load_lalonde():
    lalonde_covars = load_covars_from_path(Constants.Data.LALONDE_PATH)

    lalone_data_source = DataSource(
        covariate_data=lalonde_covars,
        binary_column_names=Constants.Data.LALONDE_DISCRETE_COVARS)

    return lalone_data_source

def load_covars_from_path(covar_path):
    '''
    Fetches covariate data from the supplied path. This function expects
    a CSV file with column names in the first row.
    '''

    return pd.read_csv(covar_path)

def load_random_normal_covariates(
    n_covars = 20, n_observations = 1000,
    mean=0, std=1, partial_correlation_degree=0.0):
    '''
    Generate random normal covariate data-frame based on the
    supplied parameters.
    '''

    covar = random_covar_matrix(
        dimension=n_covars,
        correlation_deg=partial_correlation_degree)

    # Generate random covariates
    covar_data = np.random.multivariate_normal(
        mean=np.full((n_covars,), mean),
        cov=std*covar,
        size=n_observations)

    # Normalize the generated data (this nullifies the impact of the
    # mean and std params but they are supplied for potential later use)
    covar_data = _normalize_covariate_data(covar_data)

    # Name sequentially
    covar_names = np.array([f"X{i}" for i in range(n_covars)])

    # Build DF
    norm_covars = _build_covar_data_frame(
        covar_data, covar_names, np.arange(n_observations))

    norm_data_source = DataSource(
        covariate_data=norm_covars,
        binary_column_names=[])

    return norm_data_source

def _build_covar_data_frame(data, column_names, index):
    return pd.DataFrame(
            data=data,
            columns=column_names,
            index=index)

def _normalize_covariate_data(covariate_data, exclude_columns=[]):
    included_filter = np.ones(covariate_data.shape[1])
    included_filter[exclude_columns] = 0
    excluded_filter = 1 - included_filter

    X_min = np.min(covariate_data, axis=0)
    X_max = np.max(covariate_data, axis=0)

    # Amount to shift columns. Excluded cols shift 0.
    column_shifts = (-1*X_min) * included_filter

    # Amount to scale columns. Excluded cols scale by 1.
    inverse_column_scales = ((X_max - X_min) * included_filter) + excluded_filter

    # Shift and scale to [0, 1]
    normalized_data = (covariate_data + column_shifts)/inverse_column_scales

    # Shift and scale to [-1, 1]
    rescaled_data = (normalized_data * \
        ((2*included_filter) + excluded_filter)) - \
        (1*included_filter)

    return rescaled_data



from maccabee.data_generation import DataGeneratingProcess, data_generating_method
from ..constants import Constants
from ..utilities import evaluate_expression
from ..modeling.models import CausalModel
import numpy as np
import sympy as sp
import pandas as pd

# RPY2 is used an interconnect between Python and R. It allows
# my to run R code from python which makes this experimentation
# process smoother.
import rpy2
from rpy2.robjects import IntVector, FloatVector, Formula
from rpy2.robjects.packages import importr
from rpy2.robjects import numpy2ri
numpy2ri.activate()

stats = importr('stats') # standard regression package
matching = importr('Matching') # GenMatch package

MILD_NONLINEARITY = [2]
MODERATE_NONLINEARITY = [2, 4, 7]
MILD_NONADDITIVITY = [(1,3, 0.5), (2, 4, 0.7), (4,5, 0.5), (5,6, 0.5)]
MODERATE_NONADDITIVITY = [
        (1,3, 0.5),
        (2, 4, 0.7),
        (3,5, 0.5),
        (4,6, 0.7),
        (5,7, 0.5),
        (1,6, 0.5),
        (2,3, 0.7),
        (3,4, 0.5),
        (4,5, 0.5),
        (5,6, 0.5)]

GENMATCH_SPECS = {
    "A": ([], []),
    "B": (MILD_NONLINEARITY, []),
    "C": (MODERATE_NONLINEARITY, []),
    "D": ([], MILD_NONADDITIVITY),
    "E": (MILD_NONLINEARITY, MILD_NONADDITIVITY),
    "F": ([], MODERATE_NONADDITIVITY),
    "G": (MODERATE_NONLINEARITY, MODERATE_NONADDITIVITY)
}

class GenmatchDataGeneratingProcess(DataGeneratingProcess):
    def __init__(self,
                 quadratic_indeces, interactions_list,
                 n_observations, analysis_mode):

        super().__init__(n_observations, analysis_mode)

        # Var count
        self.n_vars = 11
        self.covar_names = [f"X_{i}" for i in range(self.n_vars)]
        self.covar_symbols = np.array(sp.symbols(self.covar_names))

        # Data types (default is standard normal)
        self.binary_indeces = [1, 3, 5, 6, 8, 9]

        self.assignment_weights = np.array(
            [0, 0.8, -0.25, 0.6, -0.4, -0.8, -0.5, 0.7, 0, 0, 0])

        self.outcome_weights = np.array(
            [-3.85, 0.3, -0.36, -0.73, -0.2, 0, 0, 0, 0.71, -0.19, 0.26])
        self.true_treat_effect = -0.4

        self.quad_terms_indeces = quadratic_indeces
        self.interactions_list = np.array(interactions_list)

        self._preprocess_functions()

    def _preprocess_functions(self):
        self.treatment_logit_terms = self.assignment_weights * self.covar_symbols


         # Add quad terms
        if len(self.quad_terms_indeces) > 0:
            self.treatment_logit_terms = np.append(
                self.treatment_logit_terms, (
                    self.assignment_weights[self.quad_terms_indeces] *
                    self.covar_symbols[self.quad_terms_indeces]**2))

        # Add interact terms
        if len(self.interactions_list) > 0:
            interact_1_indeces = self.interactions_list[:, 0].astype(int)
            interact_2_indeces = self.interactions_list[:, 1].astype(int)
            interact_weights = self.interactions_list[:, 2]

            self.treatment_logit_terms = np.append(
                self.treatment_logit_terms, (
                    self.assignment_weights[interact_1_indeces] *
                       self.covar_symbols[interact_1_indeces]*
                       self.covar_symbols[interact_2_indeces]*
                       interact_weights))

        self.treatment_logit_expression = np.sum(self.treatment_logit_terms)

        self.base_outcome_terms = self.outcome_weights * self.covar_symbols
        self.base_outcome_expression = np.sum(self.base_outcome_terms)

    @data_generating_method(Constants.COVARIATES_NAME, [])
    def _generate_observed_covars(self, input_vars):
        X = np.random.normal(loc=0.0, scale=1.0, size=(
            self.n_observations, self.n_vars - 1))

        # Add bias/intercept dummy column
        X = np.hstack([np.ones((self.n_observations, 1)), X])

        # Make binary columns binary.
        for var in self.binary_indeces:
            X[:, var-1] = (X[:, var -1] > 0).astype(int)

        return pd.DataFrame(X, columns=self.covar_names)

    @data_generating_method(
        Constants.TRANSFORMED_COVARIATES_NAME,
        [Constants.COVARIATES_NAME],
        analysis_mode_only=True)
    def _generate_transformed_covars(self, input_vars):
        # Generate the values of all the transformed covariates by running the
        # original covariate data through the transforms used in the outcome and
        # treatment functions.

        observed_covariate_data = input_vars[Constants.COVARIATES_NAME]

        all_transforms = list(set(self.base_outcome_terms).union(
            self.treatment_logit_terms))

        data = {}
        for index, transform in enumerate(all_transforms):
            data[f"{Constants.TRANSFORMED_COVARIATES_NAME}{index}"] = \
                evaluate_expression(transform, observed_covariate_data)

        return pd.DataFrame(data)

    @data_generating_method(Constants.PROPENSITY_SCORE_NAME, [Constants.COVARIATES_NAME])
    def _generate_true_propensity_scores(self, input_vars):
        observed_covariate_data = input_vars[Constants.COVARIATES_NAME]

        logits = evaluate_expression(
            self.treatment_logit_expression,
            observed_covariate_data)

        return 1/(1 + np.exp(-1*logits))

    @data_generating_method(
        Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME,
        [Constants.COVARIATES_NAME])
    def _generate_outcomes_without_treatment(self, input_vars):
        observed_covariate_data = input_vars[Constants.COVARIATES_NAME]

        return evaluate_expression(
            self.base_outcome_expression,
            observed_covariate_data)

    @data_generating_method(
        Constants.TREATMENT_EFFECT_NAME,
        [Constants.COVARIATES_NAME])
    def _generate_treatment_effects(self, input_vars):
        observed_covariate_data = input_vars[Constants.COVARIATES_NAME]
        return self.true_treat_effect

class LogisticPropensityMatchingCausalModel(CausalModel):
    def __init__(self, dataset):
        self.dataset = dataset
        self.data = dataset.observed_data.drop("Y", axis=1)

    def fit(self):
        fmla = Formula('y ~ X')
        env = fmla.environment
        env['X'] = self.dataset.X.to_numpy()
        env['y'] = IntVector(self.dataset.T)

        # Run propensiy regression
        fitted_logistic = stats.glm(fmla, family="binomial")

        propensity_scores = fitted_logistic.rx2("fitted.values")

        # Run matching on prop scores
        self.match_out = matching.Match(
            Y=FloatVector(self.dataset.Y),
            Tr=IntVector(self.dataset.T),
            X=propensity_scores,
            replace=True)

    def estimate_ITE(self):
        raise NotImplementedError

    def estimate_ATE(self):
        return np.array(self.match_out.rx2("est").rx(1,1))[0]



from ..constants import Constants
from sklearn.linear_model import LinearRegression
import numpy as np


class CausalModel():
    def __init__(self, dataset):
        self.dataset = dataset

    def fit(self):
        raise NotImplementedError

    def estimate_ITE(self):
        raise NotImplementedError

    def estimate_ATE(self):
        raise NotImplementedError

    def estimate(self, estimand, *args, **kwargs):
        if estimand == Constants.Model.ATE_ESTIMAND:
            return self.estimate_ATE(*args, **kwargs)
        elif estimand == Constants.Model.ITE_ESTIMAND:
            return self.estimate_ITE(*args, **kwargs)
        else:
            raise Exception("Unrecognized estimand.")



class LinearRegressionCausalModel(CausalModel):
    def __init__(self, dataset):
        self.dataset = dataset
        self.model = LinearRegression()
        self.data = dataset.observed_data.drop("Y", axis=1)

    def fit(self):
        self.model.fit(self.data, self.dataset.Y)

    def estimate_ITE(self):
        # Generate potential outcomes
        X_under_treatment = self.data.copy()
        X_under_treatment["T"] = 1

        X_under_control = self.data.copy()
        X_under_control["T"] = 0

        y_1_predicted = self.model.predict(X_under_treatment)
        y_0_predicted = self.model.predict(X_under_control)

        ITE = y_1_predicted - y_0_predicted

        return ITE

    def estimate_ATE(self):
        # The coefficient on the treatment status
        return self.model.coef_[-1]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='maccabee',
     version='0.0.12',
     scripts=[],
     author="Josh Broomberg",
     author_email="joshbroomberg@gmail.com",
     description="Causal ML benchmarking and development tools",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/JoshBroomberg/capstone/tree/master/DataGeneration/CauseML",
     packages=setuptools.find_packages(),
     install_requires=[
          'numpy',
          'pandas',
          'scipy',
          'sympy',
          'POT',
          'pyyaml',
      ],
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
     python_requires='>=3.6',
     include_package_data=True
 )



import numpy as np
from sklearn.linear_model import LogisticRegression, LinearRegression
from scipy.spatial.distance import cdist
import ot
import pandas as pd
import matplotlib.pyplot as plt
from collections import defaultdict
from .constants import Constants
from .utilities import extract_treat_and_control_data
from .parameters import build_parameters_from_axis_levels
from .data_generation import DataGeneratingProcessSampler

### Metric functions

# These are the functions which are used
# to calculate the metric values. Each function
# is used in multiple metrics so they are named and
# parameterized generically.

def linear_regression_r2(X, y):
    if type(X) == pd.Series:
        X = X.to_numpy().reshape((-1, 1))

    lr = LinearRegression()
    lr.fit(X, y)
    return lr.score(X, y)

def logistic_regression_r2(X, y):
    if type(X) == pd.Series:
        X = X.to_numpy().reshape((-1, 1))

    lr = LogisticRegression(solver='lbfgs')
    lr.fit(X, y)
    return lr.score(X, y)

def percent(x, value):
    '''
    Percent of values in x which have the value in value.
    '''
    x = np.array(x)
    return 100*np.sum((x == value).astype(int))/len(x)

def l2_distance_between_means(covariates, treatment_status):
    '''
    L2 norm of the distance between the means of the covariates
    in the treat and control groups.
    '''
    X_treated, X_control = extract_treat_and_control_data(
        covariates, treatment_status)

    X_treated_mean = np.mean(X_treated, axis=0)
    X_control_mean = np.mean(X_control, axis=0)

    return np.linalg.norm(X_treated_mean - X_control_mean)

def mean_mahalanobis_between_nearest_counterfactual(covariates, treatment_status):
    '''
    Mahalanobis distance between the nearest neighbor of each treated unit
    which is in the control group.
    '''

    X_treated, X_control = extract_treat_and_control_data(
        covariates, treatment_status)

    # TODO: fix singular matrix issue.
    try:
        distance_matrix = cdist(X_treated, X_control, "mahalanobis")
        np.nan_to_num(distance_matrix, copy=False, nan=np.inf)
        return np.mean(np.min(distance_matrix, axis=1))
    except:
        return np.NaN

def standard_deviation_ratio(x1, x2):
    '''
    Ratio of the std of x1 and x2
    '''
    return np.std(x1)/np.std(x2)

def wasserstein(covariates, treatment_status):
    '''
    Wasserstein distance between the covariates in the treat and control groups.
    '''

    X_treated, X_control = extract_treat_and_control_data(
        covariates, treatment_status)

    num_treated, num_control = len(X_treated), len(X_control)

    a = np.ones(num_treated)/num_treated
    b = np.ones(num_control)/num_control

    M = ot.dist(X_treated, X_control)
    M /= M.max()

    # Remove all zero-cost entries to produce
    # a properly conditioned problem.
    a, b, M = ot.utils.clean_zeros(a, b, M)

    lambd = 1e-2

    wass_dist = ot.sinkhorn2(a, b, M, lambd, maxiter=3000)[0]
    if wass_dist < 1e-3 and \
        l2_distance_between_means(covariates, treatment_status) > 1e-4:
        print(f"Detected failure with W={wass_dist}...")
        return None
    else:
        return wass_dist

def naive_TE_estimate_error(TE, observed_outcome, treatment_status):
    '''
    Absolute difference between the true treatment effect and the
    naive estimate based on mean outcome in each group.
    '''
    Y_t, Y_c = extract_treat_and_control_data(observed_outcome, treatment_status)
    ATE_true = np.mean(TE)
    ATE_est = np.mean(Y_t) - np.mean(Y_c)
    return np.abs(ATE_true - ATE_est)


AnalysisMetricFunctions = Constants.AnalysisMetricFunctions
AnalysisMetricData = Constants.AnalysisMetricData
AxisNames = Constants.AxisNames

### NOTE: the code below is complex because it is designed
# to allow new metrics to be created very easily. Feel free
# to skip down to the metric definitions.

# Map the function names to function callables.
metric_functions = {
    AnalysisMetricFunctions.LINEAR_R2: linear_regression_r2,
    AnalysisMetricFunctions.LOGISTIC_R2: logistic_regression_r2,
    AnalysisMetricFunctions.PERCENT: percent,
    AnalysisMetricFunctions.L2_MEAN_DIST: l2_distance_between_means,
    AnalysisMetricFunctions.NN_CF_MAHALA_DIST: mean_mahalanobis_between_nearest_counterfactual,
    AnalysisMetricFunctions.STD_RATIO: standard_deviation_ratio,
    AnalysisMetricFunctions.WASS_DIST: wasserstein,
    AnalysisMetricFunctions.NAIVE_TE: naive_TE_estimate_error
}

# TODO: remove names and simplify access.
# Specify ways to access each input variable.
data_accessors = {
    # Covariate Data
    AnalysisMetricData.OBSERVED_COVARIATE_DATA: {
        "accessor": lambda data: data[AnalysisMetricData.OBSERVED_COVARIATE_DATA],
        "name": "X_obs"
    },
    AnalysisMetricData.TRANSFORMED_COVARIATE_DATA: {
        "accessor": lambda data: data[AnalysisMetricData.TRANSFORMED_COVARIATE_DATA],
        "name": "X_true"
    },

    # Model Variables
    Constants.OBSERVED_OUTCOME_NAME: {
        "accessor": lambda data: data[
            AnalysisMetricData.OBSERVED_OUTCOME_DATA][
            Constants.OBSERVED_OUTCOME_NAME],
        "name": Constants.OBSERVED_OUTCOME_NAME
    },

    Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME: {
        "accessor": lambda data: data[
            AnalysisMetricData.ORACLE_OUTCOME_DATA][
            Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME],
        "name": Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME
    },

    Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME: {
        "accessor": lambda data: data[
            AnalysisMetricData.ORACLE_OUTCOME_DATA][
            Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME],
        "name": Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME
    },

    Constants.TREATMENT_EFFECT_NAME: {
        "accessor": lambda data: data[
            AnalysisMetricData.ORACLE_OUTCOME_DATA][
            Constants.TREATMENT_EFFECT_NAME],
        "name": "TE"
    },

    Constants.PROPENSITY_LOGIT_NAME: {
        "accessor": lambda data: data[
            AnalysisMetricData.ORACLE_OUTCOME_DATA][
            Constants.PROPENSITY_LOGIT_NAME],
        "name": "Treat Logit"
    },

    Constants.PROPENSITY_SCORE_NAME: {
        "accessor": lambda data: data[
            AnalysisMetricData.ORACLE_OUTCOME_DATA][
            Constants.PROPENSITY_SCORE_NAME],
        "name": "P-score"
    },

    Constants.TREATMENT_ASSIGNMENT_NAME: {
        "accessor": lambda data: data[
            AnalysisMetricData.OBSERVED_OUTCOME_DATA][
            Constants.TREATMENT_ASSIGNMENT_NAME],
        "name": "T"
    }
}

# TODO: add axes for new params.

# This dict specifies the measured axes and the metrics for each axis.
# Each axis has a list of metrics which are defined
# by a metric calculation function and the arguments to be supplied to it
# as well as a unique name.
axes = {
    AxisNames.OUTCOME_NONLINEARITY: [
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.OBSERVED_COVARIATE_DATA,
                "y": Constants.OBSERVED_OUTCOME_NAME
            },
            "name": "Lin r2(X_obs, Y)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.TRANSFORMED_COVARIATE_DATA,
                "y": Constants.OBSERVED_OUTCOME_NAME
            },
            "name": "Lin r2(X_true, Y)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.OBSERVED_COVARIATE_DATA,
                "y": Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME
            },
            "name": "Lin r2(X_obs, Y1)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.OBSERVED_COVARIATE_DATA,
                "y": Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME
            },
            "name": "Lin r2(X_obs, Y0)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.TRANSFORMED_COVARIATE_DATA,
                "y": Constants.POTENTIAL_OUTCOME_WITH_TREATMENT_NAME
            },
            "name": "Lin r2(X_true, Y1)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.TRANSFORMED_COVARIATE_DATA,
                "y": Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME
            },
            "name": "Lin r2(X_true, Y0)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.OBSERVED_COVARIATE_DATA,
                "y": Constants.TREATMENT_EFFECT_NAME
            },
            "name": "Lin r2(X_obs, TE)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.TRANSFORMED_COVARIATE_DATA,
                "y": Constants.TREATMENT_EFFECT_NAME
            },
            "name": "Lin r2(X_true, TE)"
        }
    ],

    AxisNames.TREATMENT_NONLINEARITY: [
        {
            "function": AnalysisMetricFunctions.LOGISTIC_R2,
            "args": {
                "X": AnalysisMetricData.OBSERVED_COVARIATE_DATA,
                "y": Constants.TREATMENT_ASSIGNMENT_NAME
            },
            "name": "Log r2(X_obs, T)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": AnalysisMetricData.OBSERVED_COVARIATE_DATA,
                "y": Constants.PROPENSITY_LOGIT_NAME
            },
            "name": "Lin r2(X_obs, Treat Logit)"
        }
    ],

    AxisNames.PERCENT_TREATED: [
        {
            "function": AnalysisMetricFunctions.PERCENT,
            "args": {
                "x": Constants.TREATMENT_ASSIGNMENT_NAME,
                "value": 1
            },
            "name": "Percent(T==1)"
        }
    ],

    AxisNames.OVERLAP: [
        {
            "function": AnalysisMetricFunctions.NN_CF_MAHALA_DIST,
            "args": {
                "covariates": AnalysisMetricData.OBSERVED_COVARIATE_DATA,
                "treatment_status": Constants.TREATMENT_ASSIGNMENT_NAME
            },
            "name": "NN dist X_obs: T=1<->T=0"
        },
        {
            "function": AnalysisMetricFunctions.NN_CF_MAHALA_DIST,
            "args": {
                "covariates": AnalysisMetricData.TRANSFORMED_COVARIATE_DATA,
                "treatment_status": Constants.TREATMENT_ASSIGNMENT_NAME
            },
            "name": "NN dist X_true: T=1<->T=0"
        }
    ],

    AxisNames.BALANCE: [
        {
            "function": AnalysisMetricFunctions.L2_MEAN_DIST,
            "args": {
                "covariates": AnalysisMetricData.TRANSFORMED_COVARIATE_DATA,
                "treatment_status": Constants.TREATMENT_ASSIGNMENT_NAME
            },
            "name": "Mean dist X_true: T=1<->T=0"
        },
        {
            "function": AnalysisMetricFunctions.WASS_DIST,
            "args": {
                "covariates": AnalysisMetricData.TRANSFORMED_COVARIATE_DATA,
                "treatment_status": Constants.TREATMENT_ASSIGNMENT_NAME
            },
            "name": "Wass dist X_true: T=1<->T=0"
        },
        {
            "function": AnalysisMetricFunctions.WASS_DIST,
            "args": {
                "covariates": AnalysisMetricData.OBSERVED_COVARIATE_DATA,
                "treatment_status": Constants.TREATMENT_ASSIGNMENT_NAME
            },
            "name": "Wass dist X_obs: T=1<->T=0"
        },
        {
            "function": AnalysisMetricFunctions.NAIVE_TE,
            "args": {
                "TE": Constants.TREATMENT_EFFECT_NAME,
                "observed_outcome": Constants.OBSERVED_OUTCOME_NAME,
                "treatment_status": Constants.TREATMENT_ASSIGNMENT_NAME
            },
            "name": "Naive TE"
        }
    ],

    AxisNames.ALIGNMENT: [
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": Constants.OBSERVED_OUTCOME_NAME,
                "y": Constants.PROPENSITY_LOGIT_NAME
            },
            "name": "Lin r2(Y, Treat Logit)"
        },
        {
            "function": AnalysisMetricFunctions.LINEAR_R2,
            "args": {
                "X": Constants.POTENTIAL_OUTCOME_WITHOUT_TREATMENT_NAME,
                "y": Constants.PROPENSITY_LOGIT_NAME
            },
            "name": "Lin r2(Y0, Treat Logit)"
        }
    ],

    AxisNames.TE_HETEROGENEITY: [
        {
            "function": AnalysisMetricFunctions.STD_RATIO,
            "args": {
                "x1": Constants.TREATMENT_EFFECT_NAME,
                "x2": Constants.OBSERVED_OUTCOME_NAME
            },
            "name": "std(TE)/std(Y)"
        }
    ]
}

def _get_arg_value(arg_val_name, data):
    if arg_val_name in data_accessors:
        return data_accessors[arg_val_name]["accessor"](data)
    else:
        return arg_val_name

def analyze_axis_metrics_across_levels(
    axes_and_metrics, levels, data_source_generator,
    n_trials=20):
    '''
    Collect values for the given metrics and measures
    across all possible settings for each metric. Uses
    the gather_metric_measures_for_given_params to generate
    the metric measure values.
    '''

    results = defaultdict(lambda: defaultdict(dict))

    # Run for each given metric and set of measures
    for axis, metrics in axes_and_metrics.items():
        print(f"\nRunning for {axis}. Level: ", end=" ")

        # Construct observation list of measures.
        observation_spec = { axis: metrics }

        # Run trials at all levels of metric.
        for level in levels:
            print(level, end=" ")
            dgp_params = build_parameters_from_axis_levels({
                axis: level
            })

            res = gather_axis_metrics_for_given_params(
                    dgp_params, observation_spec, data_source_generator,
                    n_trials=n_trials)

            for metric, values in res[axis].items():
                results[axis][metric][level] = values

    return results

def gather_axis_metrics_for_given_params(
        dgp_params, observation_spec,
        data_source_generator,
        n_trials=10, verbose=False):

    '''
    Create n_trials datasets by sampling data generating processes according
    to the given dgp_params and data source parameters (n_covars, n_obs).

    Collect the metrics given in observation spec.
    '''
    results = defaultdict(lambda: defaultdict(list))
    for i in range(n_trials):
        if verbose:
            print("Trials run:", i+1)
        results

        # Generate data
        data_source = data_source_generator()

        # Sample DGP
        dgp_sampler = DataGeneratingProcessSampler(
            parameters=dgp_params, data_source=data_source)
        dgp = dgp_sampler.sample_dgp()
        data_set = dgp.generate_dataset()

        # Calculate metrics
        metrics = calculate_data_axis_metrics(data_set, observation_spec)

        # Build results
        for metric, measures in observation_spec.items():
            for measure in measures:
                res = metrics[metric][measure]
                if res is not None:
                    results[metric][measure].append(res)

    if verbose:
        for metric, measures in results.items():
            for measure, result_data in measures.items():
                print(f"{metric} {measure}:")
                print("min", round(np.min(result_data), 3), end=" ")
                print("mean:", round(np.mean(result_data), 3), end=" ")
                print("max", round(np.max(result_data), 3))
                print("-------------\n\n")

    return results

def calculate_data_axis_metrics(
    data_set,
    observation_spec = None):
    '''
    Given a set of generated data, calculate all metrics
    in the observation spec.
    '''

    data = {
        AnalysisMetricData.OBSERVED_COVARIATE_DATA: data_set.observed_covariate_data,
        AnalysisMetricData.OBSERVED_OUTCOME_DATA: data_set.observed_outcome_data,
        AnalysisMetricData.TRANSFORMED_COVARIATE_DATA: data_set.transformed_covariate_data,
        AnalysisMetricData.ORACLE_OUTCOME_DATA: data_set.oracle_outcome_data
    }

    axis_metric_results = {}
    for axis, axis_metrics in axes.items():
        if (observation_spec is not None) and (axis not in observation_spec):
            continue # this axis not in observation specs.

        axis_metric_results[axis] = {}

        for metric in axis_metrics:
            if (observation_spec is not None) and (metric["name"] not in observation_spec[axis]):
                 continue # this metric not in observation specs.

            func_name = metric["function"]
            func = metric_functions[func_name]

            # Assemble the argument values by fetching the relevant portions
            # of the data
            args = dict([(arg_name, _get_arg_value(arg_val_name, data))
                for arg_name, arg_val_name in metric["args"].items()])

            # Call the metric function with the args.
            res = func(**args)

            # Store results.
            metric_name = metric["name"]
            axis_metric_results[axis][f"{metric_name}"] = res

    return axis_metric_results

def plot_axis_metric_analysis(results, levels, max_measure_count=1, mean=False, min_max=False):
    '''
    Plot the results of the analyze_metric_measures_across_levels
    function. Show the median, mean and 1st and 3rd quartiles for each
    metric measure at each level.
    '''

    level_colors = ["g", "b", "y"] # low medium high will be green blue yellow
    mean_color = "r" # mean will be red
    max_min_color = "k" # min max will be black
    for metric, measures in results.items():
        plt.figure(figsize=(10, 4))
        plt.suptitle(f"{metric}")
        plt.tight_layout()

        for measure_num, (measure_name, measure_values) in enumerate(measures.items()):
            plt.subplot(1, max_measure_count, 1 + measure_num)
            plt.title(f"{measure_name}")
            for level_num, level in enumerate(levels):
                level_values = measure_values[level]

                # Find quartile values in data
                quartiles = np.percentile(
                    level_values,
                    [25, 50, 75],
                    interpolation = 'midpoint')

                # Prepare plotting data
                x = level_num+1
                y_median = quartiles[1]

                err = np.array([
                    [y_median - quartiles[0]],
                    [quartiles[2] - y_median]
                ])

                # Plot
                color = level_colors[level_num]
                plt.xlim((0, max_measure_count+1))
                plt.scatter(x, y_median, label=level, color=color)
                plt.errorbar(x, y_median, err, color=color)

                if mean:
                    y_mean = np.mean(level_values)
                    plt.scatter(x, y_mean, color=mean_color)

                if min_max:
                    y_max = np.max(level_values)
                    y_min = np.min(level_values)
                    plt.scatter(x, y_max, color=max_min_color)
                    plt.scatter(x, y_min, color=max_min_color)

            plt.legend()

        plt.show()
